# Some when did this program start?
#
# Copyright (C) 2015-2016 Andrew Cagney <cagney@gnu.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

# Everything uses this as the reference time.  There must be a more
# correct way.

from datetime import datetime


START_TIME = datetime.now()


class Lapsed:
    """A lapsed timer that prints as seconds by default

    As part of 'with' it automatically starts/stops.

    """

    def __init__(self, start=None):
        self.start = start or datetime.now()
        self.stop = None

    def format(self, now=None):
        now = now or self.stop or datetime.now()
        delta = now - self.start
        deciseconds = (delta.microseconds / 100000)
        seconds = delta.seconds % 60
        minutes = (delta.seconds // 60) % 60
        hours = (delta.seconds // 60 // 60) % 24
        days = delta.days
        if days > 0:
            # need days
            return str(delta)
        elif hours > 0:
            return "%d:%02d:%02d.%02d" % (hours, minutes, seconds,
                                          deciseconds)
        elif minutes > 0:
            return "%d:%02d.%02d" % (minutes, seconds, deciseconds)
        else:
            return "%d.%02d" % (seconds, deciseconds)

    def seconds(self, now=None):
        now = now or self.stop or datetime.now()
        delta = now - self.start
        return delta.total_seconds()

    def __enter__(self):
        self.start = datetime.now()
        return self

    def __exit__(self, type, value, traceback):
        self.stop = datetime.now()

    def __str__(self):
        return "%.01f seconds" % self.seconds()



#!/usr/bin/env python3

# Collect test statistics
#
# Copyright (C) 2015-2016 Andrew Cagney <cagney@gnu.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

from collections import defaultdict
import threading

from fab import logutil

class Counts:

    def __init__(self):
        # Use a default dict so no need to worry about initializing
        # values to zero.
        self.counts = defaultdict(lambda: defaultdict(set))
        self.lock = threading.Lock()

    def add(self, value, *keys, domain=None):
        with self.lock:
            key = "/".join(keys)
            # Include the "None" domain in the set so that it is always
            # non-empty - makes iterating over it is easier.  See
            # log_details.
            self.counts[key][value].add(domain)

    def log_summary(self, log, header=None, footer=None, prefix=""):
        with self.lock:
            if len(self.counts):
                header and log(header)
                for key, values in sorted(self.counts.items()):
                    log("%s%s: %d", prefix, key, len(values))
                footer and log(footer)

    def log_details(self, log, header=None, footer=None, prefix=""):
        with self.lock:
            if len(self.counts):
                header and log(header)
                for key, values in sorted(self.counts.items()):
                    # First invert value:domain creating domain:value
                    table = defaultdict(set)
                    for value, domains in sorted(values.items()):
                        for domain in domains:
                            table[domain].add(value)
                    # Second log key[/domain]: value ...
                    for domain, values in sorted(table.items()):
                        line = ""
                        for value in sorted(values):
                            line += " " + value
                        if domain:
                            log("%s%s/%s:%s", prefix, key, domain, line)
                        else:
                            log("%s%s:%s", prefix, key, line)
                footer and log(footer)


class Tests(Counts):
    def add(self, test, *stats):
        Counts.add(self, test.name, *stats)

# Record results:
#
# total/{passed,failed,ignored,unresolved} must add up

class Results(Counts):

    def count_result(self, result):
        Counts.add(self, result.test.name, "total")
        Counts.add(self, result.test.name, "total", str(result))
        Counts.add(self, result.test.name, "total", str(result), result.test.status)
        # details
        for issue in result.issues:
            for domain in result.issues[issue]:
                Counts.add(self, result.test.name, "total", str(result),
                           result.test.status, issue, domain=domain)

    def count_previous(self, result, previous):
        Counts.add(self, previous.test.name,
                   "status", previous.test.status,
                   result, "previous=" + str(previous))

    def add_ignored(self, test, reason):
        """The test has been excluded from the test run"""
        Counts.add(self, test.name, "total")
        Counts.add(self, test.name, "total", "ignored")
        Counts.add(self, test.name,
                   "status", test.status,
                   "ignored", reason)

    def add_skipped(self, result):
        """The test wasn't run; log the previous result"""
        self.count_result(result)
        self.count_previous("skipped", result)

    def add_result(self, result, old_result=None):
        self.count_result(result)
        if old_result:
            self.count_previous(str(result), old_result)



# Identify a test's scripts.
#
# Copyright (C) 2016 Andrew Cagney <cagney@gnu.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import os
import re

from fab.hosts import HOST_NAMES

def _scripts(directory):
    """Returns a set of *.sh scripts found in DIRECTORY"""
    scripts = set()
    if os.path.isdir(directory):
        for script in os.listdir(directory):
            if not re.match(r"[a-z0-9].*\.sh$", script):
                continue
            path = os.path.join(directory, script)
            if not os.path.isfile(path):
                continue
            # Add more filter-outs
            scripts.add(script)
    return scripts

def _add_script(run, scripts, script, host_names):
    if script in scripts:
        scripts.remove(script)
        for host_name in host_names:
            run.append((host_name, script))

def host_script_tuples(directory):
    """Return a [] list of(host, script) tuples to run"""

    scripts = _scripts(directory)

    # Form a subset of HOST_NAMES based on the names found in the
    # scripts.
    host_names = set()
    for host_name in HOST_NAMES:
        for script in scripts:
            if re.search(host_name, script):
                host_names.add(host_name)

    # init scripts: nic, east, then rest
    init_scripts = []
    _add_script(init_scripts, scripts, "nicinit.sh", ["nic"])
    _add_script(init_scripts, scripts, "eastinit.sh", ["east"])
    for host_name in sorted(host_names):
        _add_script(init_scripts, scripts, host_name + "init.sh", [host_name])

    # run scripts
    run_scripts = []
    for host_name in sorted(host_names):
        _add_script(run_scripts, scripts, host_name + "run.sh", [host_name])

    # strip out the final script
    final_scripts = []
    _add_script(final_scripts, scripts, "final.sh", sorted(host_names))

    # What's left are ordered scripts.  Preserve the order that the
    # host names appear in the file name.  For instance, the script
    # 99-west-east.sh would be run on west then east.
    extra_scripts = []
    for script in sorted(scripts):
        for host_name in re.findall("|".join(host_names), script):
            extra_scripts.append((host_name, script))

    # append the final scripts
    all_scripts = []
    all_scripts.extend(init_scripts)
    all_scripts.extend(run_scripts)
    all_scripts.extend(extra_scripts)
    all_scripts.extend(final_scripts)
    return all_scripts



# A multi-output file, for libreswan.
#
# Copyright (C) 2015 Andrew Cagney <cagney@gnu.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

def open(*streams, files=[]):
    return Tee(*streams, files=files)

class Tee:

    def __init__(self, *streams, files=[]):
        self.files = files
        self.streams = []
        for s in streams:
            self.streams.append(s)
        for f in files:
            self.streams.append(f)

    def close(self):
        for f in self.files:
            f.close()

    def write(self, text):
        for s in self.streams:
            s.write(text)

    def flush(self):
        for s in self.streams:
            s.flush()



#!/usr/bin/python
""" dist_certs.py: create a suite of x509 certificates for the Libreswan
	test harness

 Copyright (C) 2014-2015 Matt Rogers <mrogers@redhat.com>
 Copyright (C) 2015 Andrew Cagney <andrew.cagney@gmail.com>

 This program is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
 Free Software Foundation; either version 2 of the License, or (at your
 option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.

	WARNING! Your PyOpenSSL needs a patch from here:
	https://github.com/pyca/pyopenssl/pull/161
	NSS doesn't allow md5 CRL signatures. This patch lets you use
	the CRL export method and specify an acceptable signature type.

 """

import os
import sys
import ssl
import shutil
import subprocess
import time
from datetime import datetime, timedelta
import pexpect
from OpenSSL import crypto

CRL_URI = 'URI:http://nic.testing.libreswan.org/revoked.crl'

dates = {}
ca_certs = {}
end_certs = {}
endrev_name = ""
top_caname=""

def reset_files():
	for dir in ['keys/', 'cacerts/', 'certs/', 'pkcs12/',
			    'pkcs12/curveca', 'pkcs12/mainca',
				'pkcs12/otherca', 'pkcs12/badca', 'crls/']:
		if os.path.isdir(dir):
			shutil.rmtree(dir)
		os.mkdir(dir)
	for file in ['nss-pw']:
		if os.path.isfile(file):
			os.remove(file)

def writeout_cert(filename, item,
				  type=crypto.FILETYPE_PEM):
	with open(filename, "w") as f:
		f.write(crypto.dump_certificate(type, item))


def writeout_privkey(filename, item,
					 type=crypto.FILETYPE_PEM):
	with open(filename, "w") as f:
		f.write(crypto.dump_privatekey(type, item))


def create_keypair(algo=crypto.TYPE_RSA, bits=1024):
	""" Create an OpenSSL keypair
	"""
	pkey = crypto.PKey()
	pkey.generate_key(algo, bits)
	return pkey


def create_csr(pkey, CN,
			   C=None, ST=None, L=None, O=None, OU=None,
			   emailAddress=None, algo='sha1'):
	""" Create the certreq
	"""
	req = crypto.X509Req()
	subject = req.get_subject()
	subject.CN = CN
	subject.C = C
	subject.ST = ST
	subject.L = L
	subject.O = O
	subject.OU = OU
	subject.CN = CN
	subject.emailAddress = emailAddress
	req.set_pubkey(pkey)
	req.sign(pkey, algo)
	return req

def add_ext(cert, kind, crit, string):
	cert.add_extensions([crypto.X509Extension(kind, crit, string)])

def set_cert_extensions(cert, issuer, isCA=False, isRoot=False, ocsp=False, ocspuri=True):
	ku_str = 'digitalSignature'
	eku_str = ''
	ocspeku = 'serverAuth,clientAuth,codeSigning,OCSPSigning'
	cnstr = str(cert.get_subject().commonName)

	if isCA:
		ku_str = ku_str + ',keyCertSign,cRLSign'
		if "badca" in str(issuer.get_subject().commonName):
			bc = "CA:FALSE"
		else:
			bc = "CA:TRUE"
	else:
		bc = "CA:FALSE"

	add_ext(cert, 'basicConstraints', False, bc)

	if not isCA:
		SAN = "DNS: " + cnstr
		if "." in cnstr:
			ee = cnstr.split(".")[0]
			print "EE:%s"%ee
			if ee == "west" or ee == "east":
				SAN += ", email:%s@testing.libreswan.org"%ee
				if ee == "west":
					SAN += ", IP:192.1.2.45"
				else:
					SAN += ", IP:192.1.2.23"
			if ee == "otherwest" or ee == "othereast":
				SAN += ", email:%s@other.libreswan.org"%ee
		add_ext(cert, 'subjectAltName', False, SAN)

	if cnstr == 'usage-server.testing.libreswan.org':
		eku_str = 'serverAuth'
		ku_str = ku_str + ',keyEncipherment'
	elif cnstr == 'usage-client.testing.libreswan.org':
		eku_str = 'clientAuth'
		ku_str = ku_str + ',nonRepudiation'
	elif cnstr == 'usage-both.testing.libreswan.org':
		eku_str = 'serverAuth,clientAuth'
		ku_str = ku_str + ',keyEncipherment,nonRepudiation'

	if ocsp:
		ku_str = ku_str + ',keyCertSign,cRLSign'
		eku_str = ocspeku

	add_ext(cert, 'keyUsage', False, ku_str)
	if eku_str is not '':
		add_ext(cert, 'extendedKeyUsage', False, eku_str)

	if ocspuri:
		add_ext(cert, 'authorityInfoAccess', False,
		  		'OCSP;URI:http://nic.testing.libreswan.org:2560')

	add_ext(cert, 'crlDistributionPoints', False, CRL_URI)

def create_sub_cert(CN, CACert, CAkey, snum, START, END,
					C='CA', ST='Ontario', L='Toronto',
					O='Libreswan', OU='Test Department',
					emailAddress='',
					ty=crypto.TYPE_RSA, keybits=1024,
					sign_alg='sha1', isCA=False, ocsp=False):
	""" Create a subordinate cert and return the cert, key tuple
    This could be a CA for an intermediate, or not for an EE
	"""
	certkey = create_keypair(ty, keybits)
	certreq = create_csr(certkey,
						 CN, C, ST, L, O, OU,
						 emailAddress, sign_alg)

	cert = crypto.X509()
	cert.set_serial_number(snum)
	cert.set_notBefore(START)
	cert.set_notAfter(END)
	cert.set_issuer(CACert.get_subject())
	cert.set_subject(certreq.get_subject())
	cert.set_pubkey(certreq.get_pubkey())
	cert.set_version(2)

	if CN == 'nic-nourl.testing.libreswan.org':
		ocspuri = False
	else:
		ocspuri = True

	set_cert_extensions(cert, CACert, isCA=isCA, isRoot=False, ocsp=ocsp, ocspuri=ocspuri)
	cert.sign(CAkey, sign_alg)

	return cert, certkey


def create_root_ca(CN, START, END,
				   C='CA', ST='Ontario', L='Toronto',
				   O='Libreswan', OU='Test Department',
				   emailAddress='testing@libreswan.org',
				   ty=crypto.TYPE_RSA, keybits=1024,
				   sign_alg='sha1'):
	""" Create a root CA - Returns the cert, key tuple
	"""
	cakey = create_keypair(ty, keybits)
	careq = create_csr(cakey, CN, C, ST, L, O, OU,
					   emailAddress, sign_alg)

	cacert = crypto.X509()
	cacert.set_serial_number(0)
	cacert.set_notBefore(START)
	cacert.set_notAfter(END)
	cacert.set_issuer(careq.get_subject())
	cacert.set_subject(careq.get_subject())
	cacert.set_pubkey(careq.get_pubkey())
	cacert.set_version(2)

	set_cert_extensions(cacert, cacert, isCA=True, isRoot=True, ocsp=True, ocspuri=True)
	cacert.sign(cakey, sign_alg)

	return cacert, cakey


def gmc(timestamp):
	return time.strftime("%Y%m%d%H%M%SZ",
						 time.gmtime(timestamp))


def gen_gmtime_dates():
	""" Generate the dates used for this run.
	Creating openssl gmtime dates may be simpler than this.
	"""
	gmtfmt = "%b %d %H:%M:%S %Y GMT"

	ok_stamp = ssl.cert_time_to_seconds(
			time.strftime(gmtfmt, time.gmtime())) - (60*60*24)
	two_days_ago_stamp = ok_stamp - (60*60*48)
	two_days_ago_end_stamp = two_days_ago_stamp + (60*60*24)
        # Make future certs only +300 days, so we have a time overlap
        # between currently valid certs (1 year) and these futuristic certs
	future_stamp = ok_stamp + (60*60*24*365*1)
	future_end_stamp = future_stamp + (60*60*24*365*2)

	return dict(OK_NOW=gmc(ok_stamp),
				OLD=gmc(two_days_ago_stamp),
				OLD_END=gmc(two_days_ago_end_stamp),
			    FUTURE=gmc(future_stamp),
				FUTURE_END=gmc(future_end_stamp))


def store_cert_and_key(name, cert, key):
	""" Places a ca or end cert and key in the script's global store
	"""
	global ca_certs
	global end_certs

	ext = cert.get_extension(0)
	if ext.get_short_name() == 'basicConstraints':
		# compare the bytes for CA:True
		if name == "badca" or '0\x03\x01\x01\xff' == ext.get_data():
			ca_certs[name] = cert, key
		else:
			end_certs[name] = cert, key


def writeout_cert_and_key(certdir, name, cert, privkey):
	""" Write the cert and key files
	"""
	writeout_cert(certdir + name + ".crt", cert)
	writeout_privkey("keys/" + name + ".key", privkey)


def create_basic_pluto_cas(ca_names):
	""" Create the core root certs
	"""
	print "creating CA certs"
	for name in ca_names:
		print " - creating %s" % name
		ca, key = create_root_ca(CN="Libreswan test CA for " + name,
								 START=dates['OK_NOW'],
								 END=dates['FUTURE_END'])
		writeout_cert_and_key("cacerts/", name, ca, key)
		store_cert_and_key(name, ca, key)


def create_pkcs12(path, name, cert, key, ca_cert):
	""" Package and write out a .p12 file
	"""
	p12 = crypto.PKCS12()
	p12.set_certificate(cert)
	p12.set_privatekey(key)
	p12.set_friendlyname(name)
	p12.set_ca_certificates([ca_cert])
	with open(path + name + ".p12", "wb") as f:
		f.write(p12.export(passphrase="foobar"))


def create_mainca_end_certs(mainca_end_certs):
	""" Create the core set of end certs from mainca
	"""
	serial = 2
	print "creating mainca's end certs"
	for name in mainca_end_certs:
		# put special cert handling here
		print " - creating %s" % name
		if name == 'smallkey':
			keysize = 2048
		else:
			if name == 'key4096':
				keysize = 4096
			else:
				keysize = 3072

		if name == 'notyetvalid':
			startdate = dates['FUTURE']
			enddate = dates['FUTURE_END']
		elif name == 'notvalidanymore':
			startdate = dates['OLD']
			enddate = dates['OLD_END']
		else:
			startdate = dates['OK_NOW']
			enddate = dates['FUTURE_END']

		if 'other' in name:
			signer = 'otherca'
		elif name[:3] == 'bad':
			signer = 'badca'
		else:
			signer = 'mainca'

		if name == 'nic':
			ocsp_resp = True
		else:
			ocsp_resp = False

		if name == 'wrongdnorg':
			org = "No Such Agency"
		else:
			org = "Libreswan"

		if name == 'unwisechar':
			common_name = 'unwisechar ~!@#$%^&*()-'\
						  '_=+;:/?<>.testing.libreswan.org'
		elif name == 'spaceincn':
			common_name = 'space invaders.testing.libreswan.org'
		elif name == 'cnofca':
			common_name = 'Libreswan test CA for mainca'
		elif 'other' in name:
			common_name = name + '.other.libreswan.org'
		else:
			common_name = name + '.testing.libreswan.org'

		if name == 'hashsha2':
			alg = 'sha256'
		else:
			alg = 'sha1'

		if " " in common_name:
			emailAddress = "root@testing.libreswan.org"
		else:
			emailAddress = "user-%s@testing.libreswan.org"%name

		cert, key = create_sub_cert(common_name,
									ca_certs[signer][0],
									ca_certs[signer][1],
									serial, O=org,
									emailAddress=emailAddress,
									START=startdate, END=enddate,
									keybits=keysize,
									sign_alg=alg, ocsp=ocsp_resp)
		writeout_cert_and_key("certs/", name, cert, key)
		store_cert_and_key(name, cert, key)
		create_pkcs12("pkcs12/"+ signer + '/',
					  name, cert, key, ca_certs[signer][0])
		serial += 1


def create_chained_certs(chain_ca_roots, max_path, prefix=''):
	""" Create the EE->IA1->IA2->IAx-->CA chains.
	Last in the chain is the end cert
	TODO: Add more complex trust chain situations
	"""
	global endrev_name
	global top_caname
	min_path = 1
	ca_cnt = 0

	for chainca in chain_ca_roots:
		serial = len(end_certs) + ca_cnt
		lastca = ""
		#note there's an issue with the authkeyid in the chain
		#signpair = ()
		print "creating %s chain" % chainca
		for level in range(min_path, max_path):
			cname = prefix + chainca + '_int_' + str(level)

			print "level %d cname %s serial %d" % (level, cname, serial)

			if level == min_path:
				lastca = "mainca"

			signpair = ca_certs[lastca]
			print " - creating %s with the last ca of %s" % (cname, lastca)
			ca, key = create_sub_cert(cname + '.testing.libreswan.org',
									  signpair[0], signpair[1], serial,
									  START=dates['OK_NOW'],
									  END=dates['FUTURE'],
									  emailAddress="%s@testing.libreswan.org"%cname,
									  isCA=True, ocsp=False)

			writeout_cert_and_key("certs/", cname, ca, key)
			store_cert_and_key(cname, ca, key)
			lastca = cname
			serial += 1
			ca_cnt += 1

			if level == max_path - 1:
				endcert_name = prefix + chainca + "_endcert"

				signpair = ca_certs[lastca]
				print " - creating %s" % endcert_name
				ecert, ekey = create_sub_cert(endcert_name + ".testing.libreswan.org",
											  signpair[0], signpair[1], serial,
											  emailAddress="%s@testing.libreswan.org"%endcert_name,
											  START=dates['OK_NOW'],
											  END=dates['FUTURE'])

				writeout_cert_and_key("certs/", endcert_name, ecert, ekey)
				store_cert_and_key(endcert_name, ecert, ekey)
				create_pkcs12("pkcs12/", endcert_name, ecert, ekey, signpair[0])
				serial += 1

				endrev_name = prefix + chainca + "_revoked"
				top_caname = cname
				print " - creating %s" % endrev_name
				ercert, erkey = create_sub_cert(endrev_name + ".testing.libreswan.org",
											  signpair[0], signpair[1], serial,
											  emailAddress="%s@testing.libreswan.org"%endcert_name,
											  START=dates['OK_NOW'],
											  END=dates['FUTURE'])

				writeout_cert_and_key("certs/", endrev_name, ercert, erkey)
				store_cert_and_key(endrev_name, ercert, erkey)
				create_pkcs12("pkcs12/", endrev_name, ercert, erkey, signpair[0])

# this special crl was for a openswan/nss freebl combo bug, both of which should
# long be done with.

def create_leading_zero_crl():
	""" Create our special crl with a signature that starts out with '00:'
	This signs a CRL and checks for a '00' beginning. Each try increments
	the days parameter to result in a different signature
	"""
	zerosig = crypto.CRL()
	signcert, signkey = ca_certs['mainca']
	days = 1

	print "creating a CRL with a leading zero byte signature.."
	while True:
		good = False
		nl = ''

		crl = zerosig.export(signcert, signkey,
							 type=crypto.FILETYPE_TEXT, days=days, digest='sha1')
		der = zerosig.export(signcert, signkey,
							 type=crypto.FILETYPE_ASN1, days=days, digest='sha1')

		for index, line in enumerate(crl.splitlines()):
			if "Signature Algorithm" in line and index >= 5:
				nl = crl.splitlines()[index + 1].strip()
				if nl.startswith('00'):
					good = True
					break

		if good:
			print nl
			print "found after %d signatures!" % (days)
			with open("crls/crl-leading-zero-byte.crl", "wb") as f:
				f.write(der)
			break

		days += 1


def create_crlsets():
	""" Create test CRLs
	"""
	print "creating crl set"
	revoked = crypto.Revoked()
	chainrev = crypto.Revoked()
	future_revoked = crypto.Revoked()

	revoked.set_rev_date(dates['OK_NOW'])
	chainrev.set_rev_date(dates['OK_NOW'])
	future_revoked.set_rev_date(dates['FUTURE'])
	# the get_serial_number method results in a hex str like '0x17'
	# but set_serial needs a hex str like '17'
	revoked.set_serial(
			hex(end_certs['revoked'][0].get_serial_number())[2:])

	chainrev.set_serial(
			hex(end_certs['west_chain_revoked'][0].get_serial_number())[2:])

	future_revoked.set_serial(
			hex(end_certs['revoked'][0].get_serial_number())[2:])

	needupdate = crypto.CRL()
	needupdate.add_revoked(revoked)
	needupdate.add_revoked(chainrev)
	with open("crls/needupdate.crl", "wb") as f:
		f.write(needupdate.export(ca_certs['mainca'][0],
								  ca_certs['mainca'][1],
								  type=crypto.FILETYPE_ASN1,
								  days=0, digest='sha1'))

	print "sleeping for needupdate/valid crl time difference"
	time.sleep(5)
	validcrl = crypto.CRL()
	validcrl.add_revoked(revoked)
	validcrl.add_revoked(chainrev)
	with open("crls/cacrlvalid.crl", "wb") as f:
		f.write(validcrl.export(ca_certs['mainca'][0],
								ca_certs['mainca'][1],
								type=crypto.FILETYPE_ASN1,
								days=15, digest='sha1'))

	othercrl = crypto.CRL()
	othercrl.add_revoked(revoked)
	othercrl.add_revoked(chainrev)
	with open("crls/othercacrl.crl", "wb") as f:
		f.write(othercrl.export(ca_certs['otherca'][0],
								ca_certs['otherca'][1],
								type=crypto.FILETYPE_ASN1,
								days=15, digest='sha1'))

	notyet = crypto.CRL()
	notyet.add_revoked(future_revoked)
	with open("crls/futurerevoke.crl", "wb") as f:
		f.write(notyet.export(ca_certs['mainca'][0],
							  ca_certs['mainca'][1],
							  type=crypto.FILETYPE_ASN1,
							  days=15, digest='sha1'))

	 #create_leading_zero_crl()


def create_ec_certs():
	""" The OpenSSL module doesn't appear to have
	support for curves so we do it with pexpect
	"""
	print "creating EC certs"
	#create CA
	pexpect.run('openssl ecparam -out keys/curveca.key '
				'-name secp384r1 -genkey -noout')
	child = pexpect.spawn('openssl req -x509 '
						  '-new -key keys/curveca.key '
						  '-out cacerts/curveca.crt '
						  '-days 3650 -set_serial 1')
	child.expect('Country Name')
	child.sendline('CA')
	child.expect('State')
	child.sendline('Ontario')
	child.expect('Locality')
	child.sendline('Toronto')
	child.expect('Organization')
	child.sendline('Libreswan')
	child.expect('Organizational')
	child.sendline('Test Department')
	child.expect('Common')
	child.sendline('Libreswan test EC CA')
	child.expect('Email')
	child.sendline('testing@libreswan.org')
	child.expect(pexpect.EOF)

	serial = 2
	for name in ['east', 'west', 'north', 'road']:
		print "- creating %s-ec" % name
		#create end certs
		pexpect.run('openssl ecparam -out keys/' + name +
					'-ec.key -name secp384r1 -genkey -noout')
		child = pexpect.spawn('openssl req -x509 '
							  '-new -key keys/curveca.key '
							  '-out certs/' + name +
							  '-ec.crt -days 365 -set_serial ' +
							  str(serial))
		child.expect('Country Name')
		child.sendline('CA')
		child.expect('State')
		child.sendline('Ontario')
		child.expect('Locality')
		child.sendline('Toronto')
		child.expect('Organization')
		child.sendline('Libreswan')
		child.expect('Organizational')
		child.sendline('Test Department')
		child.expect('Common')
		child.sendline(name + '-ec.testing.libreswan.org')
		child.expect('Email')
		child.sendline('testing@libreswan.org')
		child.expect(pexpect.EOF)
		serial += 1
		#package p12
		pexpect.run('openssl pkcs12 -export '
					'-inkey keys/%s-ec.key '
				    '-in certs/%s-ec.crt -name %s-ec '
					'-certfile cacerts/curveca.crt '
					'-caname "curveca" '
					'-out pkcs12/curveca/%s-ec.p12 '
					'-passin pass:foobar -passout pass:foobar'
					% (name, name, name, name))


def run_dist_certs():
	""" Generate the pluto test harness x509
	certificates, p12 files, keys, and CRLs
	"""
	# Add root CAs here
	basic_pluto_cas =  ('mainca', 'otherca', 'badca')
	# Add end certs here
	mainca_end_certs = ('nic','east','west', 'road', 'sunset',
						'sunrise','north','south',
						'pole','park','beet','carrot',
					    'usage-server', 'usage-client',
					    'usage-both',
						'nic-noext', 'nic-nourl',
						'japan','bigkey', 'key4096',
						'notyetvalid','notvalidanymore',
						'signedbyother','otherwest','othereast','wrongdnorg',
						'unwisechar','spaceincn','hashsha2',
						'cnofca','revoked', 'badwest', 'badeast')
	# Add chain roots here
	chain_ca_roots =   ('east_chain', 'west_chain')

	# Put special case code for new certs in the following functions
	create_basic_pluto_cas(basic_pluto_cas)
	create_mainca_end_certs(mainca_end_certs)
	create_chained_certs(chain_ca_roots, 3)
	create_chained_certs(chain_ca_roots, 9, 'long_')
	create_chained_certs(chain_ca_roots, 10, 'too_long_')
	create_crlsets()
	create_ec_certs()

def create_nss_pw():
	print "creating nss-pw"
	f = open("nss-pw","w")
	f.write("foobar")
	f.close()

def main():
	outdir = os.path.dirname(sys.argv[0])
	cwd = os.getcwd()
	os.chdir(outdir)
	global dates
	reset_files()
	dates = gen_gmtime_dates()
	print "format dates being used for this run:"
	# TODO: print the display GMT times
	for n, s in dates.iteritems():
		print "%s : %s" % (n, s)

	run_dist_certs()

	create_nss_pw()
	os.chdir(cwd)
	print "finished!"


if __name__ == "__main__":
	main()



# Some json parsing functions.
#
# Copyright (C) 2016 Andrew Cagney <cagney@gnu.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import json
from datetime import datetime

class result:
    directory = "directory"
    testname = "testname"
    expect = "expect"
    result = "result"
    hosts = "hosts"
    issues = "errors" # for historic reasons, "issues" are publicly called "errors"
    script_time = "script_time"
    boot_time = "boot_time"
    total_time = "total_time"
    hosts = "hosts"
    # keep things going
    time = "time"
    runtime = "runtime"


class results:
    summary = "summary"
    table = "table"

class summary:
    passed = "passed"
    failed = "failed"
    incomplete = "incomplete"
    good = "good"
    total = "Total"
    # end-time: YYYY-MM-DD HH:MM see ftime/ptime.
    date = "date"
    runtime = "runtime"
    directory = "directory"

def load(io):
    try:
        return json.load(io)
    except ValueError:
        return None

def loads(s):
    try:
        return json.loads(s)
    except ValueError:
        return None

def default(obj):
    if hasattr(obj, "json") and callable(getattr(obj, "json")):
        return obj.json()
    if hasattr(obj, "isoformat") and callable(getattr(obj, "isoformat")):
        # date/time objects
        return obj.isoformat()
    elif hasattr(obj, "__str__") and callable(getattr(obj, "__str__")):
        return str(obj)
    else:
        print("obj:", obj)
        raise TypeError(obj)

def dump(j, io):
    json.dump(j, io, indent=2, default=default)

def dumps(s):
    return json.dumps(s, default=default)

def ftime(t):
    """Format the time"""
    return t.strftime("%Y-%m-%d %H:%M")

def ptime(s):
    """Tries to parse the time"""
    # print(s)
    for hms in ["%H:%M", "%H:%M:%S", "%H:%M:%S.%f"]:
        for t in [" ", "T"]:
            try:
                t = datetime.strptime(s, "%Y-%m-%d" + t + hms)
                # print(t)
                return t
            except:
                None
    return None



# Perform post.mortem on a test result, for libreswan.
#
# Copyright (C) 2015-2016 Andrew Cagney <cagney@gnu.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import os
import re
import subprocess
import difflib
import weakref
import gzip
import bz2

from fab import logutil

# Strings used to mark up files; see also runner.py where it marks up
# the file names.
CUT = ">>>>>>>>>>cut>>>>>>>>>>"
TUC = "<<<<<<<<<<tuc<<<<<<<<<<"
DONE = CUT + " done " + TUC


class Resolution:
    PASSED = "passed"
    FAILED = "failed"
    UNRESOLVED = "unresolved"
    UNTESTED = "untested"
    UNSUPPORTED = "unsupported"
    def __init__(self):
        self.state = None
    def __str__(self):
        return self.state
    def __eq__(self, rhs):
        return self.state == rhs
    def isresolved(self):
        return self.state in [self.PASSED, self.FAILED]
    def unsupported(self):
        assert(self.state in [None])
        self.state = self.UNSUPPORTED
    def untested(self):
        assert(self.state in [None])
        self.state = self.UNTESTED
    def passed(self):
        assert(self.state in [None])
        self.state = self.PASSED
    def failed(self):
        assert(self.state in [self.FAILED, self.PASSED, self.UNRESOLVED])
        if self.state in [self.FAILED, self.PASSED]:
            self.state = self.FAILED
    def unresolved(self):
        assert(self.state in [self.PASSED, self.FAILED, self.UNRESOLVED, None])
        self.state = self.UNRESOLVED


# Mapping between hosts and their issues and/or issues and their
# hosts.
#
# Two maps are maintained:
#
# - the ISSUE_HOSTS map is indexed by ISSUE, each ISSUE entry then
#   contains a list (set?) of hosts
#
#   This is so that code can easily determine if a specific issue,
#   regardless of the HOST, has occurred.  All the programatic
#   operators, such as __contains__(), are implemented based on this
#   model.
#
#   XXX: Could probably re-implement issues as an extension of
#   dictionary.
#
# - the HOST_ISSUES map is indexed by HOST, each HOST entry then
#   contains a list (set?) of issues
#
#   This is used to display and dump the issues (__str__(), json()).
#   It seems that the most user friendly format is:
#   host:issue,... host:issue:,...

class Issues:

    ASSERTION = "ASSERTION"
    EXPECTATION = "EXPECTATION"

    CORE = "CORE"
    SEGFAULT = "SEGFAULT"
    GPFAULT = "GPFAULT"

    CRASHED = {ASSERTION, EXPECTATION, CORE, SEGFAULT, GPFAULT}

    OUTPUT_MISSING = "output-missing"
    OUTPUT_UNCHECKED = "output-unchecked"
    OUTPUT_TRUNCATED = "output-truncated"
    OUTPUT_WHITESPACE = "output-whitespace"
    OUTPUT_DIFFERENT = "output-different"

    ABSENT = "absent"

    SANITIZER_FAILED = "sanitizer-failed"

    BASELINE_FAILED = "baseline-failed"
    BASELINE_PASSED = "baseline-passed"
    BASELINE_MISSING = "baseline-missing"
    BASELINE_WHITESPACE = "baseline-whitespace"
    BASELINE_DIFFERENT = "baseline-different"

    def __init__(self, logger):
        # Structure needs to be JSON friendly.
        self._host_issues = {}
        self._issue_hosts = {}
        self._logger = logger

    # Both __str__() and json() dump the table in user friendly
    # format.  That is:
    #
    #     host:issue,...  host:issue,...
    #
    # This is the opposite to what code expects - a dictionary
    # structured issue:host,... .

    def __str__(self):
        s = ""
        for host, errors in sorted(self._host_issues.items()):
            if s:
                s += " "
            if host:
                s += host + ":"
            s += ",".join(sorted(errors))
        return s

    def json(self):
        return self._host_issues

    # Programatic collections like interface.  This is indexed by
    # ISSUE so that it is easy to query Issues to see if an ISSUE
    # occurred on any host.

    def __bool__(self):
        return len(self._issue_hosts) > 0

    def __iter__(self):
        return self._issue_hosts.keys().__iter__()

    def __contains__(self, issue):
        return issue in self._issue_hosts

    def __getitem__(self, issue):
        return self._issue_hosts[issue]

    def add(self, issue, host):
        if not host in self._host_issues:
            self._host_issues[host] = []
        if not issue in self._host_issues[host]:
            self._host_issues[host].append(issue)
        if not issue in self._issue_hosts:
            self._issue_hosts[issue] = []
        if not host in self._issue_hosts[issue]:
            self._issue_hosts[issue].append(host)
        self._logger.debug("host %s has issue %s", host, issue)


def _strip(s):
    s = re.sub(r"[ \t]+", r"", s)
    s = re.sub(r"\n+", r"\n", s)
    s = re.sub(r"^\n", r"", s)
    return s

def _whitespace(l, r):
    """Return true if L and R are the same after stripping white space"""
    return _strip(l) == _strip(r)

def _diff(logger, ln, l, rn, r):
    """Return the difference between two strings"""

    if l == r:
        # slightly faster path
        logger.debug("_diff '%s' and '%s' fast match", ln, rn)
        return []
    # compare
    diff = list(difflib.unified_diff(l.splitlines(), r.splitlines(),
                                     fromfile=ln, tofile=rn,
                                     lineterm=""))
    logger.debug("_diff: %s", diff)
    if not diff:
        # Always return a list.
        return []
    return diff


def _sanitize_output(logger, raw_path, test):
    # Run the sanitizer found next to the test_sanitize_directory.
    command = [
        test.testing_directory("utils", "sanitizer.sh"),
        raw_path,
        test.testing_directory("pluto", test.name)
    ]
    logger.debug("sanitize command: %s", command)
    # Note: It is faster to re-read the file than read the
    # pre-loaded raw console output.
    process = subprocess.Popen(command, stdin=subprocess.DEVNULL,
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    logger.debug("sanitized output:\n%s", stdout)
    if process.returncode or stderr:
        # any hint of an error
        logger.error("sanitize command '%s' failed; exit code %s; stderr: '%s'",
                     command, process.returncode, stderr.decode("utf8"))
        return None
    return stdout.decode("utf-8")


def _load_file(logger, filename):
    """Load the specified file; return None if it does not exist"""

    if os.path.exists(filename):
        logger.debug("loading file: %s", filename)
        with open(filename) as f:
            return f.read()
    else:
        logger.debug("file %s does not exist", filename)
    return None


# The TestResult objects are almost, but not quite, an enum. It
# carries around additional result details.

class TestResult:

    def __str__(self):
        return str(self.resolution)

    def __bool__(self):
        """True if the test was attempted.

        That is POSIX 1003.3 PASS, FAIL, or UNRESOLVED (which leaves
        UNTESTED).

        """
        return self.resolution in [self.resolution.PASSED, self.resolution.FAILED, self.resolution.UNRESOLVED]

    def __init__(self, logger, test, quick, update=None,
                 output_directory=None):

        # Set things up for passed
        self.logger = logger
        self.test = test
        self.resolution = Resolution()
        self.issues = Issues(self.logger)
        self.diffs = {}
        self.sanitized_output = {}
        self.grub_cache = {}
        self.output_directory = output_directory or test.output_directory

        # If there is no OUTPUT directory the result is UNTESTED -
        # presence of the OUTPUT is a clear indicator that some
        # attempt was made to run the test.
        if not os.path.exists(self.output_directory):
            self.resolution.untested()
            self.logger.debug("output directory missing: %s", self.output_directory)
            return

        # Start out assuming that it passed and then prove otherwise.
        self.resolution.passed()

        # did pluto crash?
        for host_name in test.host_names:
            pluto_log_filename = host_name + ".pluto.log"
            if self.grub(pluto_log_filename, "ASSERTION FAILED"):
                self.issues.add(Issues.ASSERTION, host_name)
                self.resolution.failed()
            if self.grub(pluto_log_filename, "EXPECTATION FAILED"):
                self.issues.add(Issues.EXPECTATION, host_name)
                # XXX: allow expection failures?

        # Check the raw console output for problems and that it
        # matches expected output.
        for host_name in test.host_names:

            # Check that the host's raw output is present.
            #
            # If there is no output at all then the test crashed badly
            # (for instance, while trying to boot domains).
            #
            # Since things really screwed up, mark the test as
            # UNRESOLVED and give up.

            raw_output_filename = host_name + ".console.verbose.txt"
            if self.grub(raw_output_filename) is None:
                self.issues.add(Issues.OUTPUT_MISSING, host_name)
                self.resolution.unresolved()
                # With no raw console output, there's little point in
                # trying validating it.  Skip remaining tests for this
                # host.
                continue

            # Check the host's raw output for signs of a crash.

            self.logger.debug("host %s checking raw console output for signs of a crash",
                              host_name)
            if self.grub(raw_output_filename, r"[\r\n]CORE FOUND"):
                self.issues.add(Issues.CORE, host_name)
                self.resolution.failed()
            if self.grub(raw_output_filename, r"SEGFAULT"):
                self.issues.add(Issues.SEGFAULT, host_name)
                self.resolution.failed()
            if self.grub(raw_output_filename, r"GPFAULT"):
                self.issues.add(Issues.GPFAULT, host_name)
                self.resolution.failed()

            # Check that the host's raw output is complete.
            #
            # The output can become truncated for several reasons: the
            # test is a work-in-progress; it crashed due to a timeout;
            # or it is still being run.
            #
            # Don't try to match the prompt ("#").  If this is the
            # first command in the script, the prompt will not appear
            # in the output.
            #
            # When this happens, mark it as a FAIL.  The
            # test-in-progress case was hopefully handled further back
            # with the RESULT hack forcing the test to UNRESOLVED.

            ending = ": ==== end ===="
            logger.debug("host %s checking if raw console output contains '%s' (or '%s')",
                         host_name, DONE, ending)
            if self.grub(raw_output_filename, DONE) is None \
            and self.grub(raw_output_filename, ending) is None:
                # this is probably truncated output; but if the test
                # is old it may not be the case. and an unresolved
                # test, but need to first exclude other options.
                self.issues.add(Issues.OUTPUT_TRUNCATED, host_name)
                if os.path.isfile(test.result_file(self.output_directory)):
                    self.resolution.failed()
                else:
                    self.resolution.unresolved()

            # Sanitize what ever output there is and save it.
            #
            # Even when the output is seemingly truncated this is
            # useful.

            sanitized_output_path = os.path.join(self.output_directory,
                                                 host_name + ".console.txt")
            self.logger.debug("host %s sanitize console output '%s'",
                              host_name, sanitized_output_path)
            sanitized_output = None
            if quick:
                sanitized_output = _load_file(self.logger,
                                              sanitized_output_path)
            if sanitized_output is None:
                sanitized_output = _sanitize_output(self.logger,
                                                    os.path.join(self.output_directory, raw_output_filename),
                                                    test)
            if sanitized_output is None:
                self.issues.add(Issues.SANITIZER_FAILED, host_name)
                self.resolution.unresolved()
                continue
            if update:
                self.logger.debug("host %s updating sanitized output file: %s",
                                  host_name, sanitized_output_path)
                with open(sanitized_output_path, "w") as f:
                    f.write(sanitized_output)

            self.sanitized_output[host_name] = sanitized_output

            expected_output_path = test.testing_directory("pluto", test.name,
                                                          host_name + ".console.txt")
            self.logger.debug("host %s comparing against known-good output '%s'",
                              host_name, expected_output_path)

            expected_output = _load_file(self.logger, expected_output_path)
            if expected_output is None:
                self.issues.add(Issues.OUTPUT_UNCHECKED, host_name)
                self.resolution.unresolved()
                continue

            diff = None
            diff_filename = host_name + ".console.diff"

            if quick:
                # Try to load the existing diff file.  Like _diff()
                # save a list of lines.
                diff = self.grub(diff_filename)
                if diff is not None:
                    diff = diff.splitlines()

            if diff is None:
                # use brute force
                diff = _diff(self.logger,
                             "MASTER/" + test.directory + "/" + host_name + ".console.txt",
                             expected_output,
                             "OUTPUT/" + test.directory + "/" + host_name + ".console.txt",
                             sanitized_output)

            if update:
                self.logger.debug("host %s updating diff file %s",
                                  host_name, diff_filename)
                # Always create the diff file; when there is no diff
                # leave it empty.
                with open(os.path.join(self.output_directory, diff_filename), "w") as f:
                    if diff:
                        for line in diff:
                            f.write(line)
                            f.write("\n")

            if diff:
                self.diffs[host_name] = diff
                whitespace = _whitespace(expected_output,
                                         sanitized_output)
                self.resolution.failed()
                if whitespace:
                    self.issues.add(Issues.OUTPUT_WHITESPACE, host_name)
                else:
                    self.issues.add(Issues.OUTPUT_DIFFERENT, host_name)

    def grub(self, filename, regex=None, cast=lambda x: x):
        """Grub around FILENAME to find regex"""
        self.logger.debug("grubbing '%s' for '%s'", filename, regex)
        # Find/load the file, and uncompress when needed.
        if not filename in self.grub_cache:
            self.grub_cache[filename] = None
            for suffix, open_op in [("", open), (".gz", gzip.open), (".bz2", bz2.open),]:
                path = os.path.join(self.output_directory, filename + suffix)
                if os.path.isfile(path):
                    self.logger.debug("loading '%s' into cache", path)
                    with open_op(path, "rt") as f:
                        self.grub_cache[filename] = f.read()
                        break
        contents = self.grub_cache[filename]
        if contents is None:
            return None
        if regex is None:
            return contents
        match = re.search(regex, contents, re.MULTILINE)
        if not match:
            return None
        group = match.group(len(match.groups()))
        self.logger.debug("grub '%s' matched '%s'", regex, group)
        return cast(group)


# XXX: given that most of args are passed in unchagned, this should
# change to some type of result factory.

def mortem(test, args, domain_prefix="",
           baseline=None, output_directory=None,
           quick=False, update=False):

    logger = logutil.getLogger(domain_prefix, __name__, test.name)

    test_result = TestResult(logger, test, quick,
                             output_directory=output_directory,
                             update=update)

    if not test_result:
        return test_result

    if not baseline:
        return test_result

    # For "baseline", the general idea is that "kvmresults.py | grep
    # baseline" should print something when either a regression or
    # progression has occurred.  For instance:
    #
    #    - a test passing but the baseline failing
    #
    #    - a test failing, but the baseline passing
    #
    #    - a test failing, and the baseline failling in a different way
    #
    # What isn't interesting is a test and the baseline failing the
    # same way.

    if not test.name in baseline:
        test_result.issues.add(Issues.ABSENT, "baseline")
        return test_result

    # When loading the baseline results use "quick" so that the
    # original results are used.  This seems to be the best of a bad
    # bunch.
    #
    # Since that the baseline was generated using an old sanitizer and
    # reference output, using the latest sanitizer scripts (in
    # testing/) can, confusingly, lead to baselines results being
    # identified as failures failing yet the diffs show a pass.
    #
    # OTOH, when this goes to compare the results against the
    # baseline, first putting them through the latest sanitizer tends
    # to result in better diffs.

    base = baseline[test.name]
    baseline_result = TestResult(logger, base, quick=True)

    if not baseline_result.resolution in [test_result.resolution.PASSED,
                                          test_result.resolution.FAILED]:
        test_result.issues.add(str(baseline_result), "baseline")
        return test_result

    if test_result.resolution in [test_result.resolution.PASSED] \
    and baseline_result.resolution in [baseline_result.resolution.PASSED]:
        return test_result

    for host_name in test.host_names:

        # result missing output; still check baseline ..
        if host_name not in test_result.sanitized_output:
            if host_name in baseline_result.sanitized_output:
                if host_name in baseline_result.diffs:
                    test_result.issues.add(Issues.BASELINE_FAILED, host_name)
                else:
                    test_result.issues.add(Issues.BASELINE_PASSED, host_name)
            continue

        if not host_name in baseline_result.sanitized_output:
            test_result.issues.add(Issues.BASELINE_MISSING, host_name)
            continue

        if not host_name in test_result.diffs:
            if host_name in baseline_result.diffs:
                test_result.issues.add(Issues.BASELINE_FAILED, host_name)
            continue

        if not host_name in baseline_result.diffs:
            test_result.issues.add(Issues.BASELINE_PASSED, host_name)
            continue

        baseline_diff = _diff(logger,
                              "BASELINE/" + test.directory + "/" + host_name + ".console.txt",
                              baseline_result.sanitized_output[host_name],
                              "OUTPUT/" + test.directory + "/" + host_name + ".console.txt",
                              test_result.sanitized_output[host_name])
        if baseline_diff:
            baseline_whitespace = _whitespace(baseline_result.sanitized_output[host_name],
                                              test_result.sanitized_output[host_name])
            if baseline_whitespace:
                test_result.issues.add(Issues.BASELINE_WHITESPACE, host_name)
            else:
                test_result.issues.add(Issues.BASELINE_DIFFERENT, host_name)
            # update the diff to something hopefully closer?
            # test_result.diffs[host_name] = baseline_diff
        # else:
        #    test_result.issues.add("baseline-failed", host_name)

    return test_result



#!/usr/bin/python2 -tt
# -*- coding: utf-8 -*-

"""Is connection encrypted?

./is_encrypted.py --debug --port 443 10.6.6.1
./is_encrypted.py --debug --source 2001:998:2e:f00::42 2a00:1190:c00a:f00::1
./is_encrypted.py --debug --source 193.65.3.113 172.27.129.42

/*
 * Is connection encrypted? -utility.
 *
 * Copyright (C) 2018  Kim B. Heino <b@bbbs.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
"""

from __future__ import print_function
from __future__ import unicode_literals
import argparse
import swan


def main():
    """Main program"""
    parser = argparse.ArgumentParser(
        description='check if traffic would get encrypted by IPsec')
    parser.add_argument('--debug', action='store_true', help='show debugging')
    parser.add_argument('--source', help='source address of the packet')
    parser.add_argument('--port', help='connect to TCP port on destination')
    parser.add_argument('destination', help='destination IP address to check')
    args = parser.parse_args()

    print(swan.is_encrypted(
        args.destination, args.port, args.source, debug=args.debug))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
import sys
import time
import os
import subprocess
import re
import argparse
import logging
from fab import shell
from fab import argutil

try:
    import pexpect
    import setproctitle
except ImportError as e:
    module = str(e)[16:]
    sys.exit("we require the python module %s " % module)

# XXX: This function's behaviour should not depend on the presence of
# FILENAME.
def read_exec_shell_cmd(child, filename, timer):
    if os.path.exists(filename):
        f_cmds = None
        try:
            f_cmds = open(filename, "r")
            for line in f_cmds:
                line = line.strip()
                # We need the lines with # for the cut --- tuc
                # sections if line and not line[0] == '#':
                if line:
                    child.run(line, timeout=timer)
        finally:
            if f_cmds:
                f_cmds.close
    else:
        child.run(filename, timeout=timer)

def connect_to_kvm(args):

    vmlist = subprocess.getoutput("sudo virsh list")
    running = 0
    for line in vmlist.split("\n")[2:]:
       try:
            num,host,state = line.split()
            if host == args.hostname and state == "running":
               running = 1
               print("Found %s running already" % args.hostname)
               continue
       except:
               pass

    if args.reboot:
       waittime = 20
       if not running:
            print("Booting %s - pausing %s seconds" % (args.hostname,waittime))
            subprocess.getoutput("sudo virsh start %s" % args.hostname)
            time.sleep(waittime)
       else:
            subprocess.getoutput("sudo virsh reboot %s" % args.hostname)
            print("Rebooting %s - pausing %s seconds" % (args.hostname,waittime))
            time.sleep(waittime)

    print("Taking %s console by force" % args.hostname)
    cmd = "sudo virsh console --force %s" % args.hostname
    timer = 120
    # Need to use spawnu with python3.
    child = shell.Remote(cmd, hostname=args.hostname, username="root")
    child.output(sys.stdout)
    prompt = child.prompt
    print("Shell prompt '%s'" % prompt.pattern)

    done = 0
    tries = 60
    print("Waiting on %s login: or shell prompt" % (args.hostname))
    while not done and tries != 0:
      try:
        print("sending ctrl-c return")
        #child = pexpect.spawn (cmd)
        #child.sendcontrol('c')
        child.sendline ('')
        print("found, waiting on login: or shell prompt")
        res = child.expect (['login: ', prompt], timeout=3)
        if res == 0:
           print("sending login name root")
           child.sendline ('root')
           print("found, expecting password prompt")
           child.expect ('Password:', timeout=1)
           print("found, sending password")
           child.sendline ('swan')
           print("waiting on root shell prompt")
           child.expect (prompt, timeout=1)
           print("done")
           done = 1
        elif res == 1:
          print('----------------------------------------------------')
          print('Already logged in as root on %s' % args.hostname)
          print('----------------------------------------------------')
          done = 1
      except:
        print("(%s [%s] waiting)" % (args.hostname,tries))
        tries -= 1
        time.sleep(1)

    if not done:
        print('console is not answering on host %s, aborting'%args.hostname)
        return None

    # Make certain that both ends are in sync and set up as expected.
    child.sync()

    return child

def run_final (args, child):
    timer = 30
    output_file = "./OUTPUT/%s.console.verbose.txt" % (args.hostname)
    f = open(output_file, 'a')
    child.logfile = f
    cmd = "./final.sh"
    if os.path.exists(cmd):
        read_exec_shell_cmd(child, cmd, timer)
    f.close
    return

def compile_on(args, child):
    child.chdir(args.sourcedir)
    timer = 900
    cmd = "%s/testing/guestbin/swan-build" % (args.sourcedir)
    status = child.run(cmd, timeout=timer)
    if status:
        sys.exit(status)

def make_install(args, child):
    child.chdir(args.sourcedir)
    timer=300
    cmd = "%s/testing/guestbin/swan-install" % (args.sourcedir)
    status = child.run(cmd, timeout=timer)
    if status:
        sys.exit(status)

def run_test(args, child):
    #print 'HOST : ', args.hostname
    #print 'TEST : ', args.testname

    timer = 120
    child.chdir("%s/pluto/%s " % (args.testdir, args.testname))

    output_file = "./OUTPUT/%s.console.verbose.txt" % (args.hostname)
    f = open(output_file, 'w')
    child.logfile = f

    # do we need to prep x509?
    if args.x509:
	# call to runkvm.py forced it
        x509 = "--x509"
    else:
        x509 = ""

    cmd = "./%sinit.sh" %  (args.hostname)
    read_exec_shell_cmd(child, cmd, timer)

    cmd = "./%srun.sh" %  (args.hostname)
    if os.path.exists(cmd):
        read_exec_shell_cmd(child, cmd, timer)
        run_final(args,child)
        f.close
    else:
        f.close

def main():
    setproctitle.setproctitle("swankvm")
    parser = argparse.ArgumentParser(description='runkvm arguments.')
    parser.add_argument('--testname', '-t', action='store', help='The name of the test to run.')
    parser.add_argument('--hostname', '-H', action='store', default='', help='The name of the host to run.')
    parser.add_argument('--compile', action="store_true", help='compile the source on host <hostname>.')
    parser.add_argument('--install', action="store_true", help='run make install module_install .')
    parser.add_argument('--x509', action="store_true", help='tell the guest to setup the X509 certs in NSS.')
    parser.add_argument('--final', action="store_true", help='run final.sh on the host.')
    parser.add_argument('--reboot', action="store_true", help='first reboot the host')
    parser.add_argument('--sourcedir', action='store', default='/source', help='source <directory> to build')
    parser.add_argument('--testdir', action='store', default='/testing', help='test <directory> to run tests from')
    parser.add_argument('--runtime', type=argutil.timeout, default=None,
                        help='max run-time (timeout) for the run command (default infinite)')
    parser.add_argument("--log-level", default="warning", metavar="LEVEL",
                        help="Set logging level to %(metavar)s (default: %(default)s)")
    # unused parser.add_argument('--timer', default=120, help='timeout for each command for expect.')
    parser.add_argument('domain', action='store', nargs='?', help="name of the domain to connect to")
    parser.add_argument('command', action='store', nargs='?', help="command to run")
    args = parser.parse_args()
    # HACK: Merge the hostname arguments
    args.hostname = args.hostname or args.domain
    if not args.hostname:
        print("Either hostname or domain must be specified")
        sys.exit(1)

    logging.basicConfig(level=args.log_level.upper())

    child = connect_to_kvm(args)

    if not child:
        sys.exit("Failed to launch/connect to %s - aborted" % args.hostname)

    if args.command:
        child.chdir(args.sourcedir)
        status = child.run(args.command, timeout=args.runtime)
        sys.exit(status)

    if args.compile:
        compile_on(args, child)

    if args.install:
        make_install(args, child)

    if (args.testname and not args.final):
        run_test(args,child)

    if args.final:
        run_final(args,child)

    # if there's nothing else to do, create an interactive shell.
    batch_mode = args.command or args.compile or args.install or args.testname or args.final or args.reboot
    if not batch_mode:

        print()
        print()
        child.output(None)
        print("Escape character is ^]")
        # Hack so that the prompt appears
        child.output(sys.stdout)
        child.run("")
        child.output(None)
        # Normal mode
        child.stty_sane()
        child.interact()

if __name__ == "__main__":
    main()



# tcpdump driver, for libreswan
#
# Copyright (C) 2016 Andrew Cagney <cagney@gnu.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import re
import subprocess
from os import path

from fab import logutil


def _interfaces(domains):
    interfaces = set()
    for domain in domains:
        status, output = domain.dumpxml()
        if status:
            domain.logger.error("dumpxml failed: %s" % (output))
            continue
        bridges = re.compile(r"\<source network='[^']*' bridge='([^']*)'/\>").findall(output)
        if bridges:
            for bridge in bridges:
                interfaces.add(bridge)
            continue
        networks = re.compile(r"\<source network='([^']*)'/\>").findall(output)
        if networks:
            for network in networks:
                interfaces.add(network)
            continue
        domain.logger.error("domain has no interfaces")
    return interfaces


class Dump:

    def __init__(self, logger, domain_prefix, output_directory, domains, enable=True):
        self.enable = enable
        self.logger = logger
        self.domains = domains
        self.domain_prefix = domain_prefix
        self.output_directory = output_directory
        self.tcpdumps = []

    def __enter__(self):
        if self.enable:
            interfaces = _interfaces(self.domains)
            self.logger.info("tcpdump interfaces: %s", interfaces)
            for interface in list(interfaces):
                # strip the prefix
                _,_,log = interface.partition(self.domain_prefix)
                log = path.join(self.output_directory, log)
                command = [
                    "sudo", "tcpdump",
                    "-i", interface,
                    "-w", log + ".pcap",
                    "-s", "0", "-n", "not stp and not port 22"
                ]
                self.logger.info("tcpdump command: %s", command)
                with open(log + ".log", "w") as out:
                    t = subprocess.Popen(command, stdin=subprocess.DEVNULL,
                                         stdout=out, stderr=subprocess.STDOUT,
                                         start_new_session=True)
                    self.tcpdumps.append(t)
        self.logger.info("tcpdump processes: %s",
                         " ".join([str(t.pid) for t in self.tcpdumps]))

    def __exit__(self, type, value, traceback):
        for t in self.tcpdumps:
            kill = ["sudo", "/usr/bin/kill", str(t.pid)]
            self.logger.info("killing tcpdump with: %s", kill)
            for attempt in range(5):
                k = subprocess.Popen(kill, stdin=subprocess.DEVNULL)
                output, error = k.communicate()
                self.logger.info("%s output: %d '%s' '%s'", kill, k.returncode, output, error)
                try:
                    t.wait(timeout=1)
                    self.logger.info("exit code %d", t.returncode)
                    break
                except subprocess.TimeoutExpired:
                    self.logger.info("timedout %s", str(t.pid))



# Stuff to talk to virsh, for libreswan
#
# Copyright (C) 2015-2016 Andrew Cagney <cagney@gnu.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import re
import os
import logging
import pexpect
import time

from fab import virsh
from fab import shell
from fab import timing
from fab import logutil

MOUNTS = {}

def mounts(domain):
    """Return a table of 9p mounts for the given domain"""
    # maintain a local cache
    if domain in MOUNTS:
        domain.logger.debug("using mounts from cache")
        return MOUNTS[domain]
    mounts = {}
    MOUNTS[domain] = mounts
    status, output = domain.dumpxml()
    if status:
        # XXX: Throw an exception?
        return mounts
    for line in output.splitlines():
        if re.compile("<filesystem type='mount' ").search(line):
            source = ""
            target = ""
            continue
        match = re.compile("<source dir='([^']*)'").search(line)
        if match:
            source = match.group(1)
            # Strip trailing "/" along with other potential quirks
            # such as the mount point being a soft link.
            source = os.path.realpath(source)
            continue
        match = re.compile("<target dir='([^']*)'").search(line)
        if match:
            target = match.group(1)
            continue
        if re.compile("<\/filesystem>").search(line):
            domain.logger.debug("filesystem target '%s' source '%s'", target, source)
            mounts[target] = source
            continue
    return mounts


FSTABS = {}

def mount_point(domain, console, device):
    """Find the mount-point for device"""
    if not domain in FSTABS:
        FSTABS[domain] = {}
    fstab = FSTABS[domain]
    if device in fstab:
        mount= fstab[device]
        domain.logger.debug("using fstab entry for %s (%s) from cache", device, mount)
        return mount;
    console.sendline("awk '$1==\"" + device + "\" { print $2 }' < /etc/fstab")
    status, match = console.expect_prompt("(/\S+)")
    mount = match.group(1)
    fstab[device] = mount
    domain.logger.debug("fstab has device '%s' mounted on '%s'", device, mount)
    return mount


def directory(domain, console, directory, default=None):
    directory = os.path.realpath(directory)
    for target, source in mounts(domain).items():
        if os.path.commonprefix([source, directory]) == source:
            # found a suitable mount point, now find were it is
            # mounted on the remote machine.
            root = mount_point(domain, console, target)
            return root + directory[len(source):]
    return default


# Domain timeouts

SHUTDOWN_TIMEOUT = 20
START_TIMEOUT = 10
LOGIN_PROMPT_TIMEOUT = 120

def _wait_for_login_prompt(domain, console, timeout, also_expect=[]):
    # If match is non-empty, append it, so the first index is 1
    timer = timing.Lapsed()
    # Create a new list, otherwise the default [] ends up containing
    # lots of login prompts ...
    matches = ["login: "] + also_expect
    domain.logger.debug("waiting %d seconds for %s",
                        timeout, " or ".join(ascii(match) for match in matches))
    match = console.expect(matches, timeout=timeout)
    # always report a prompt match, let caller decide if other cases
    # should be verbose.
    domain.logger.log(match == 0 and logutil.INFO or logutil.DEBUG,
                      "%s matched after %s", ascii(matches[match]), timer)
    return match


# Assuming the machine is booted, try to log-in.

LOGIN_TIMEOUT = 10
PASSWORD_TIMEOUT = 5
SHELL_TIMEOUT = 5

def _login(domain, console, username, password, login_timeout, password_timeout, shell_timeout):

    lapsed_time = timing.Lapsed()

    domain.logger.info("waiting %s seconds for login prompt; %s seconds for password prompt; %s seconds for shell prompt",
                       login_timeout, password_timeout, shell_timeout)
    domain.logger.debug("console prompt: %s", console.prompt.pattern)

    domain.logger.debug("sending control-c+carriage return, waiting %s seconds for login or shell prompt", login_timeout)
    console.sendintr()
    console.sendline("")
    if _wait_for_login_prompt(domain, console, timeout=login_timeout,
                              also_expect=[console.prompt]) == 1:
        # shell prompt
        domain.logger.info("We're in after %s!  Someone forgot to log out ...", lapsed_time)
        return

    domain.logger.debug("sending username '%s' waiting %s seconds for password or shell prompt", \
                        username, password_timeout)
    console.sendline(username)
    if console.expect(["Password: ", console.prompt], timeout=password_timeout):
        # shell prompt
        domain.logger.info("We're in after %s! No password ...", lapsed_time)
        return

    domain.logger.debug("sending password '%s', waiting %s seconds for shell prompt",
                        password, shell_timeout)
    console.sendline(password)
    console.expect(console.prompt, timeout=shell_timeout)
    domain.logger.info("We're in after %s!", lapsed_time)


# The machine is assumed to be booted.

def login(domain, console,
          username="root", password="swan",
          login_timeout=LOGIN_TIMEOUT,
          password_timeout=PASSWORD_TIMEOUT,
          shell_timeout=SHELL_TIMEOUT):
    if not console:
        domain.logger.error("domain not running")
        return None
    # try to login
    _login(domain, console, username=username, password=password,
           login_timeout=login_timeout,
           password_timeout=password_timeout,
           shell_timeout=shell_timeout)
    console.sync()


# Get a domain running with an attatched console.  Should be really
# quick.

def _start(domain, timeout):
    domain.logger.info("starting domain")
    # Bring the machine up from scratch.
    end_time = time.time() + timeout
    # Do not call this when the console is functional!
    console = domain.console()
    if console:
        raise pexpect.TIMEOUT("console should not be open")
    first_attempt = True
    while console == None:
        if time.time() > end_time:
            pexpect.TIMEOUT("Trying to get a console")
        status, output = domain.start()
        if status and first_attempt:
            # The first attempt at starting the domain _must_ succeed.
            # Failing is a sign that the domain was running.  Further
            # attempts might fail as things get racey.
            raise pexpect.TIMEOUT("failed to start domain: %s" % output)
        time.sleep(1)
        # try opening the console again
        console = domain.console()
        first_attempt = False
    domain.logger.debug("got console")
    return console


# Use the console to detect the shutdown - if/when the domain stops it
# will exit giving an EOF.

def shutdown(domain, console=None, shutdown_timeout=SHUTDOWN_TIMEOUT):
    lapsed_time = timing.Lapsed()
    console = console or domain.console()
    if not console:
        domain.logger.error("domain already shutdown")
        return None
    domain.logger.info("waiting %d seconds for domain to shutdown", shutdown_timeout)
    domain.shutdown()
    if console.expect([pexpect.EOF,pexpect.TIMEOUT], timeout=shutdown_timeout):
        domain.logger.error("timeout waiting for shutdown, destroying it")
        domain.destroy()
        if console.expect([pexpect.EOF,pexpect.TIMEOUT], timeout=shutdown_timeout):
            domain.logger.error("timeout waiting for destroy, giving up")
            return True
        return False
    domain.logger.info("domain shutdown after %s", lapsed_time)
    return False

def _reboot_to_login_prompt(domain, console):

    # Drain any existing output.
    console.drain()

    # The reboot pattern needs to match all the output up to the point
    # where the machine is reset.  That way, the next pattern below
    # can detect that the reset did something and the machine is
    # probably rebooting.
    timeouts = [SHUTDOWN_TIMEOUT, START_TIMEOUT, LOGIN_PROMPT_TIMEOUT]
    timeout = 0
    for t in timeouts:
        timeout += t
    domain.logger.info("waiting %s seconds for reboot and login prompt", timeout)
    domain.reboot()
    timer = timing.Lapsed()
    for timeout in timeouts:
        # pexpect's pattern matcher is buggy and, if there is too much
        # output, it may not match "reboot".  virsh's behaviour is
        # also buggy, see further down.
        match = _wait_for_login_prompt(domain, console, timeout=timeout,
                                       also_expect=["reboot: Power down\r\n",
                                                    pexpect.EOF,
                                                    pexpect.TIMEOUT])
        if match == 0:
            return console
        elif match == 1:
            domain.logger.info("domain rebooted after %s", timer)
        elif match == 2:
            # On F26, in response to the reset(?), virsh will
            # spontaneously disconnect.
            domain.logger.error("domain disconnected spontaneously after %s", timer)
            break
        elif match == 3 and console.child.buffer == "":
            # On F23, F24, F25, instead of resetting, the domain will
            # hang.  The symptoms are a .TIMEOUT and an empty buffer
            # (HACK!).
            domain.logger.error("domain appears stuck, no output received after waiting %d seconds",
                                timeout)
            break

    # Things aren't going well.  Per above Fedora can screw up or the
    # domain is just really slow.  Try destroying the domain and then
    # cold booting it.

    destroy = True
    if domain.state() == virsh.STATE.PAUSED:
        destroy = False
        domain.logger.error("domain suspended, trying resume")
        status, output = domain.resume()
        if status:
            domain.logger.error("domain resume failed: %s", output)
            destroy = True
    if destroy:
        domain.logger.error("domain hung, trying to pull the power cord")
        domain.destroy()
        console.expect_exact(pexpect.EOF, timeout=SHUTDOWN_TIMEOUT)
        console = _start(domain, timeout=START_TIMEOUT)

    # Now wait for login prompt.  If this second attempt fails then
    # either a .TIMEOUT or a .EOF exception will be thrown and the
    # test will be aborted (marked as unresolved).
    _wait_for_login_prompt(domain, console, timeout=LOGIN_PROMPT_TIMEOUT)
    return console


def boot_to_login_prompt(domain, console):

    if console:
        return _reboot_to_login_prompt(domain, console)
    else:
        console = _start(domain, timeout=START_TIMEOUT)
        _wait_for_login_prompt(domain, console, timeout=LOGIN_PROMPT_TIMEOUT)
        return console



#!/usr/bin/python
#
# example-terminate.py
#
# This is a script that can be ADDED to the updown scripts. Normally this would
# be addd to /usr/libexec/ipsec/_updown.netkey (for protostack=netkey)
# This is an example that logs XAUTH users to /tmp/terminate.log when the
# tunnel goes down:
#
# Inside _updown.netkey add a call to this script, like:
#
#
#    down-client)
#        # connection to my client subnet going down
#        downrule
#        # If you are doing a custom version, firewall commands go here.
#        restoreresolvconf
#        /usr/local/sbin/example-terminate.py $PLUTO_XAUTH_USERNAME
#

import os, sys, pipes
import commands
import requests
import signal

def child(username):
	# Send a GET request to our logserver to log disconnecting client
	# this might block if unavailable, so we must be detached from
	# pluto.
	url = 'https://logserver.example.com/delete/'
	payload = {'username': username}
	r = requests.get(url, data=payload)

	fp = open("/tmp/terminate.log","a")
	fp.write("Termination notification sent for %s\n"%username)
	fp.write("HTTP status code:%s\n"%r.status_code)
	fp.write(r.text)
	fp.write("\n")
	fp.close()
	os._exit(0)

if __name__ == '__main__':
	if len(sys.argv) < 2:
		# nothing to log
		sys.exit(0)

	username = sys.argv[1]
	if username != pipes.quote(username):
		sys.exit("bogus characters in username '%s', ignored termination request"%username)

	# Redirect standard file descriptors to ensure pluto does not block on us
	os.close(0)
	os.close(1)
	os.close(2)
	os.close(3) # pluto.log - workaround for bug #202

	# Do a double fork to decouple from the parent environment
	pid = os.fork()
	if pid > 0:
		# exit first parent
		sys.exit(0)

	os.chdir("/")
	os.setsid()
	os.umask(0)

	# do second fork
	pid = os.fork()
	if pid > 0:
		# exit from second parent
		sys.exit(0)

	# we're fully detached now - take as long as needed to do the work
	child(username)

